﻿/* 
    Copyright (c) 2012 Microsoft Corporation.  All rights reserved.
    Use of this sample source code is subject to the terms of the Microsoft license 
    agreement under which you licensed this sample source code and is provided AS-IS.
    If you did not accept the terms of the license agreement, you are not authorized 
    to use this sample source code.  For the terms of the license, please see the 
    license agreement between you and Microsoft.
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media.Imaging;
using Microsoft.FaceSdk.Alignment;
using Microsoft.FaceSdk.Common;
using Microsoft.FaceSdk.Common.Extension;
using Microsoft.FaceSdk.Image;
using Microsoft.FaceSdk.ImageHelper;
using Microsoft.Phone.Controls;
using UIFramework.Controls;
using SdkColor = Microsoft.FaceSdk.Image.Color;
using SdkImage = Microsoft.FaceSdk.Image.Image<byte>;
using FaceParty.Exception;
using System.Threading;
using Microsoft.Phone.Tasks;
using System.IO;

namespace FaceParty
{
    public partial class EditPage : PhoneApplicationPage
    {
        // Scale the face rectangle given by the detector a bit so as to cover the entire the face region
        private const double HorizontalScaleOfFaceRect = 1.2;
        private const double VerticalScaleOfFaceRect = 1.5;

        // TODO: fast face detecting configuration, to be removed when release;
        //private const double HorizontalScaleOfFaceRect = 1.1;
        //private const double VerticalScaleOfFaceRect = 1.2;

        //nark the detection Failed
        private bool IsDetectionFailed { get; set; }

        /// <summary>
        /// Faces in the current photo
        /// </summary>
        private FaceWrapper[] currentFaces;

        /// <summary>
        /// X/Y scale of the photo on the image control
        /// </summary>
        private ImageTransform transform = new ImageTransform();

        /// <summary>
        /// The preview image as shown on the screen, which may be much smaller than the actual image.
        /// The image processing algorithms actually works on this image to save computational cost.
        /// </summary>
        private SdkImage rawImage;
        private WriteableBitmap rawBitmap;

        /// <summary>
        /// Thread for detecting face(s) in the given image
        /// </summary>
        /// <param name="parameter">Thread parameter, should be Image<byte></param>
        private void WorkThread(object parameter)
        {
            Image<byte> image = parameter as Image<byte>;

            Debug.Assert(image != null);
            if (image == null)
            {
                throw new InvalidOperationException("Invalid type for work thread parameter");
            }

            this.faceDetectingAnimation.DetectionBegin();
            this.LayoutRoot.IsHitTestVisible = false;

            Dispatcher.BeginInvoke(() =>
            {
                currentFaces = Detect(image);

                if (this.currentFaces.Length == 0)
                {
                    this.faceDetectingAnimation.DetectionFail();

                    //messageBox for FailDetection and back to the mainpage
                    String errormsg = null;

                    //if the current page has back to the mainPage
                    //the messagebox should be Prohibited
                    if (NavigationService.CurrentSource.OriginalString != Localization.Strings.EditPageName)
                        return;

                    errormsg = Localization.Strings.noneFaceDetectionState;

                    if (MessageBoxResult.OK == MessageBox.Show(errormsg))
                    {
                        MainPage.IsBackFromFaceDetectionFailed = true;
                        NavigationService.GoBack();
                        return;
                    }
                }
                else
                {
                    this.faceDetectingAnimation.DetectionSuccess();
                }

                this.LayoutRoot.IsHitTestVisible = true;

                foreach (FaceWrapper f in this.currentFaces)
                {
                    var mask = CreateMask(f);
                    this.maskCanvas.Children.Add(mask);
                }
            });
        }

        /// <summary>
        /// Detect faces in the given image
        /// </summary>
        private static FaceWrapper[] Detect(Image<byte> image)
        {
            FaceWrapper[] faceWrappers = null;

            lock (FaceParty.Common.FaceFinder.SyncRoot)
            {
                var faces = FaceParty.Common.FaceFinder.Current.DetectFace(image);
                faceWrappers = (from f in faces select new FaceWrapper { Face = f }).ToArray();
            }

            return faceWrappers;
        }

#if DEBUG
        private void DisplayAlignment(Image<byte> image)
        {
            // display face alignments for debugging
            Dispatcher.BeginInvoke(() =>
            {
                using (var render = Renderer.Create(new ImageARGB(image)))
                {
                    foreach (var f in this.currentFaces)
                    {
                        render.DrawPoints(f.Face.Shape, 3, SdkColor.Red);
                        render.DrawRectangle(SdkColor.Green, 2, (RectangleF)f.Face.Region);
                    }
                    this.imgViewer.Source = ImageConverter.SdkToSystem(render.Image);
                }
            });
        }
#endif

        /// <summary>
        /// Calculate the inplane rotation angle (in degree) of a face.
        /// </summary>
        /// <param name="faceShape">ASM face shape</param>
        /// <returns>Rotation angle in degree</returns>
        public static float GetFaceInplaneRotationAngle(IList<PointF> faceShape)
        {
            // We estimate the rotation angle based on the line linking the centers of the two eyes
            var leftEyeConf = ASMModelConf.Lookup(ASMFaceComponent.LeftEye);
            var rightEyeConf = ASMModelConf.Lookup(ASMFaceComponent.RightEye);

            var leftEyeCenter = Enumerable.Range(leftEyeConf.Index, leftEyeConf.Count).Select(i => faceShape[i]).AroundRect().Center;
            var rightEyeCenter = Enumerable.Range(rightEyeConf.Index, rightEyeConf.Count).Select(i => faceShape[i]).AroundRect().Center;

            var eyeLine = leftEyeCenter - rightEyeCenter;

            var angle = System.Math.Atan2(eyeLine.Y, eyeLine.X);
            var rotation = (angle < 0 ? (System.Math.PI + angle) / 2 : (angle - System.Math.PI) / 2) * 180 / System.Math.PI;

            return (float)rotation;
        }
    }
}
