﻿/* 
    Copyright (c) 2012 Microsoft Corporation.  All rights reserved.
    Use of this sample source code is subject to the terms of the Microsoft license 
    agreement under which you licensed this sample source code and is provided AS-IS.
    If you did not accept the terms of the license agreement, you are not authorized 
    to use this sample source code.  For the terms of the license, please see the 
    license agreement between you and Microsoft.
*/

using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.IO.IsolatedStorage;
using System.Threading;

using Microsoft.FaceSdk.Common;
using Microsoft.FaceSdk.Common.Extension;
using Microsoft.FaceSdk.Image;
using Microsoft.FaceSdk.Alignment;
using Microsoft.FaceSdk.Detection;
using System.Windows;
using Size = Microsoft.FaceSdk.Common.Size;

namespace FaceParty.Common
{
    public class Face
    {
        public Rectangle Region { get; set; }
        public PointF[] Shape { get; set; }
        public PointF[] Forehead { get; set; }
    }

    public class FaceFinder
    {
        #region Private fields

        public static readonly object SyncRoot = new object();

        private IFaceDetector _detector;
        private IFaceAlignmentor _alignmentor;

        private const int MaxImageDimension = 600;
        private const int MinFaceSize = 48;
        private const float DetectorStep = 0.11f;

        #endregion
        #region Constructors

        static FaceFinder()
        {
            ThreadPool.QueueUserWorkItem((WaitCallback)FaceFinder.Create);
        }

        private FaceFinder()
        {
            this.Initialize();
        }

        private static void Create(object obj)
        {
            if (Current == null)
            {
                lock (FaceFinder.SyncRoot)
                {
                    if (Current == null)
                    {
                        Current = new FaceFinder();
                    }
                }
            }
        }

        #endregion
        #region Public methods

        /// <summary>
        /// Detect faces in the given image and detect the feature points
        /// </summary>
        /// <param name="image">Input image</param>
        /// <returns>Detected faces with feature points</returns>
        public Face[] DetectFace(Image<byte> image)
        {
            if (image == null) throw new ArgumentNullException("image", "Invalid parameter.");
            if (image.Channels != Color.GrayChannels && image.Channels != Color.ArgbChannels)
                throw new InvalidOperationException("Color image required.");

            var smallImg = image;
            float scale = 1.0f;
            if (smallImg.Width > MaxImageDimension || smallImg.Height > MaxImageDimension)
            {
                smallImg = CropScale.ScaleImage(image, MaxImageDimension, out scale);
            }

            var gray = new ImageGray(smallImg);
            FaceRect[] faces = _detector.Detect(gray, new Size(MinFaceSize, MinFaceSize), gray.Size, DetectorStep, true, null, smallImg);

            List<Face> alignedFaces = new List<Face>(faces.Length);
            float invScale = 1.0f / scale;
            foreach (var f in faces)
            {
                var shape = _alignmentor.Align(gray, f.Rect);
                if (shape != null)
                {
                    shape.UpdateEach(p => p * invScale);
                    alignedFaces.Add(
                        new Face
                        {
                            Shape = shape,
                            Region = (Rectangle)(f.Rect * invScale),
                            Forehead = InterpolateForehead(shape)
                        });
                }
            }

            return alignedFaces.ToArray();
        }

        public static PointF[] InterpolateForehead(IList<PointF> faceShape)
        {
            // interop fore head
            PointF[] foreheadShape = new PointF[5];

            foreheadShape[2] = faceShape[36] + faceShape[47] - faceShape[57];
            //foreheadShape[2] = faceShape[37].Add(faceShape[46]).Substract(faceShape[51]);
            foreheadShape[0] = faceShape[68] - (faceShape[69] - faceShape[68]) / 2.0f;
            foreheadShape[4] = faceShape[86] - (faceShape[85] - faceShape[86]) / 2.0f;

            var pt0 = (foreheadShape[0] + foreheadShape[4]) / 2.0f;
            var pt1 = (foreheadShape[0] + foreheadShape[2]) / 2.0f;
            var pt2 = (foreheadShape[2] + foreheadShape[4]) / 2.0f;

            foreheadShape[1] = pt1 - (pt0 - pt1) / 2.0f;
            foreheadShape[3] = pt2 - (pt0 - pt2) / 2.0f;

            return foreheadShape;
        }

        #endregion
        #region Public properties

        public static FaceFinder Current { get; private set; }

        #endregion
        #region Private methods

        private void Initialize()
        {
            this._detector = FaceDetectorFactory.Create(FaceDetectionType.MultiviewPyramid);
            this._alignmentor = FaceAlignmentorFactory.Create(FaceAlignmentType.Asm);
        }

        #endregion
    }
}
